/***************************************************************************
                          downloader.cpp  -  description
                             -------------------
    begin                : Thu Jun 24 2004
    copyright            : (C) 2004 by David Pye
    email                : dmp@davidmpye.dyndns.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <qdir.h>
#include <qfile.h>
#include <qtimer.h>
#include <qdatastream.h>
#include <qsocket.h>
#include <qguardedptr.h>

#include <kdebug.h>

#include "downloader.h"
#include "server.h"
#include "file.h"
#include "nntppost.h"

Downloader::Downloader(Server *server, QString tempDir, QString postPrefix) {
	m_tempDir = tempDir;
	m_server = server;                                                   
	m_postPrefix = postPrefix;

	m_post = NULL;
	
	//Connect to server, and await instructions.
	//We will need to delete this socket manually, once we have 'escaped' its signals.
	//So, don't parent it, otherwise the Qt event loop may delete it inconveniently.
	m_socket = new QSocket(this);
	
	connect(m_socket, SIGNAL(connected()),this,SLOT(socketConnectedOk()));
	connect(m_socket, SIGNAL(readyRead()),this,SLOT(dataReady()));
	connect(m_socket, SIGNAL(error(int)),this,SLOT(socketError(int)));
  	
	m_timeoutTimer = new QTimer(this);
	connect(m_timeoutTimer, SIGNAL(timeout()), this, SLOT(timeOut()));
	m_timeoutTimer->start(m_server->timeout()*1000, TRUE );
	
	m_speedCalcTimer = new QTimer(this);
	connect(m_speedCalcTimer,SIGNAL(timeout()),this,SLOT(speedCalc()));

	m_loggedIn=false;
	
	//Connect to the server
	m_state = INITIAL_CONNECT;

	m_socket->connectToHost(m_server->hostName(),m_server->port());
}

Downloader::~Downloader(){
	//If we are connected, disconnect.
	if (m_socket->state()==QSocket::Connected) {
		writeBlock("QUIT\r\n");
		m_socket->flush();
		m_socket->close();

	}
}

void Downloader::socketConnectedOk() {
}

void Downloader::downloadPost(NntpPost *tmppost) {
	m_post = tmppost;

	if (m_state==READY) {
		m_speedCalcTimer->start( 1000, false );
		m_timeoutTimer->start( m_server->timeout()*1000, TRUE );

		for (int i=0; i<5; ++i) m_speeds[i]=0;

		m_previousBytesDown=0;
		m_bytesDown=0;
		m_speedCount=0;
		m_speedRate=0;
		m_previousAnnouncedPercentage=0;
	
		//kdDebug()<<"Beginning download for "<<tmppost->title()<<"\n";
		if (m_currentGroup!=m_post->group()) changeGroup();
		else requestPost();
	}
}

void Downloader::timeOut() {
	kdDebug()<<"Downloader::A timeout occurred!\n";
	m_timeoutTimer->stop();
	m_speedCalcTimer->stop();
	
	emit error(this,ERR_TIMEOUT);
}

void Downloader::speedCalc() {
	//This isnt that nice, but we average the speed out over the previous five values
	if (m_speedCount==4) m_speedCount=0;
	m_speeds[m_speedCount] = m_bytesDown-m_previousBytesDown;

	int speedAvg=0;
	for (int i=0; i<5; ++i) {
		speedAvg+=m_speeds[i];
	}
	speedAvg/=5;
	emit progress(this,SPEED,speedAvg/1024);
	m_previousBytesDown = m_bytesDown;
	m_speedCount++;
}

void Downloader::dataReady() {
	//Reset the timeout timer
	m_timeoutTimer->start( m_server->timeout()*1000, TRUE );
	//If the data isn't newline terminated, we don't want to know yet
	while (m_socket->canReadLine()) {
		QString line = m_socket->readLine();
		//Is this the initial connection message?

		switch (m_state) {
			
    		case INITIAL_CONNECT:
			if (line.left(1)=="2") {
				m_state = READY;
				emit message(this,CONNECTED);
				if (m_post) downloadPost(m_post);
			}
			else {
				kdDebug()<<"Server error at connect - server said "<<line<<"\n";
			}
			break;

		case CHANGING_GROUP:
			if (line.left(1)=="2") {
				m_currentGroup = m_post->group();
				emit message(this,CHANGED_GROUP);
				requestPost();
			}
			else if (line.left(1)=="4") {
				//Let's try auth-ing.
				if (!m_loggedIn && m_server->authRequired()) {
					loginUsername();
				}
				else {
					kdDebug()<<"Unable to change group - server said "<<line<<"\n";
					emit error(this, ERR_NO_GROUP);
				}
			}
    			break;

		case AUTH_USERNAME:
			if (line.left(1)=="3") {
				//Username accepted, please send password
				loginPassword();
			}
			break;

		case AUTH_PASSWORD:
			if (line.left(1)=="2") {
				m_loggedIn = true;
				emit message(this, AUTHED);
				changeGroup();
			}
			else {
				//Login failed.
				kdDebug()<<"Login failed - server said "<<line<<"\n";
				emit error(this,ERR_AUTH);
			}
			break;
			
		case REQUESTING_POST:
			if (line.left(1)=="2") {
				//Looks good - the post will come down next
				m_state = DOWNLOADING_POST;
  	    			m_postFile.setName(m_tempDir+"/"+m_postPrefix+m_post->title());
				m_postFile.open(IO_WriteOnly|IO_Truncate);
				m_postStream.setDevice(&m_postFile);
				emit message(this, DOWNLOADING);
				emit progress(this,PERCENT,0);
				//This eats the newline that always comes next
				m_socket->readLine();
			}
			else {
				//Post isn't available
				kdDebug()<<"Post is unavailable - server said "<<line<<"\n";
				m_state=READY;
				//We are in a ready state - we can handle another post request.
				emit error(this, ERR_NO_POST);
			}
			break;
			
		case DOWNLOADING_POST:
			int percent;
			if (line.at(0)=='.') {
				if (line.at(1)!='.') {
		//			kdDebug()<<"Downloader: Completed post "+m_post->title()+"\n";
					m_state = READY;
					m_postFile.close();
					m_postStream.unsetDevice();
					m_timeoutTimer->stop();
					m_speedCalcTimer->stop();
          				emit progress(this,PERCENT,100);
					emit message(this, COMPLETED);
					break;
				}
				else line.remove(0,1);
			}
			m_postStream.writeRawBytes(line.data(),line.length());
			m_bytesDown+=(line.length())-4;
			percent = (m_bytesDown*100) / m_post->bytes();
			//We only want to announce every 5 percent
			if (percent>m_previousAnnouncedPercentage+4) {
        			m_previousAnnouncedPercentage = percent;
				emit progress(this,PERCENT,percent);
			}
			break;

		default:
			kdDebug()<<"Received message "<<line<<" while in READY state\n";
		}
	}
}

void Downloader::changeGroup() {
	m_state = CHANGING_GROUP;
	//kdDebug()<<"Downloader: Changing to group "<<m_post->group()<<"\n";
	writeBlock("GROUP "+m_post->group()+"\r\n");
}

void Downloader::loginUsername() {
	m_state = AUTH_USERNAME;
	//kdDebug()<<"Downloader:: Supplying username "+m_server->username()<<"\n";
	writeBlock("AUTHINFO USER "+m_server->username()+"\r\n");	
}

void Downloader::loginPassword() {
	m_state = AUTH_PASSWORD;
	//kdDebug()<<"Downloader:: Supplying password (not shown)"<<"\n";
	writeBlock("AUTHINFO PASS "+m_server->password()+"\r\n");	
}

void Downloader::requestPost() {
	emit message(this,REQUESTING);
	m_state=REQUESTING_POST;
	//kdDebug()<<"Downloader: requesting post "<<m_post->title()<<"\n";
	writeBlock("BODY <"+m_post->title()+">\r\n");
}

int Downloader::writeBlock(QString string) {
	return m_socket->writeBlock(string.ascii(), string.length());
}

void Downloader::socketError(int err) {
	kdDebug()<<"Downloader: socket error occurred\n";
	switch (err) {
		case QSocket::ErrHostNotFound:
			emit error(this, ERR_DNS);
			break;
		case QSocket::ErrConnectionRefused:
			emit error(this, ERR_CONN_REFUSED);
			break;
		default:
			kdDebug()<<"Unhandled socket error\n";
	}
}

NntpPost* Downloader::post() const {
	return m_post;
}

Server* Downloader::server() const {
	return m_server;
}
