/***************************************************************************
		 downloadmanager.cpp  -  The manager for a .nzb file download
                             -------------------
    begin                : Thu Jun 24 2004
    copyright            : (C) 2004 by David Pye
    email                : dmp@davidmpye.dyndns.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qfile.h>
#include <qstringlist.h>
#include <qptrlist.h>
#include <qdir.h>
#include <qvaluelist.h>
#include <qtimer.h>

#include <kapplication.h>
#include <klocale.h>
#include <kdebug.h>

#include "downloadmanager.h"
#include "downloader.h"
#include "server.h"
#include "servermanager.h"
#include "file.h"
#include "nntppost.h"
#include "decoder.h"
#include "yydecoder.h"
#include "uudecoder.h"

DownloadManager::DownloadManager(){
	m_serverManager=NULL;
	m_currentFile = NULL;
	m_downloadInProgress=false;
	m_postPrefix = "knzb-"+QString::number(getpid())+"-";
	m_serverManager = new ServerManager();
}

DownloadManager::~DownloadManager(){
	//Delete any temporary files left in the temp dir from this knzb run.
	QDir d(m_tempDir);
	QStringList files = d.entryList(m_postPrefix+"*");
	QStringList::Iterator end = files.end();
	for (QStringList::Iterator it = files.begin(); it!=end; it++) {
		d.remove(*it);
	}
}

void DownloadManager::applySettings(Settings *s) {
	m_tempDir = s->tempDir;
	m_destDir = s->destDir;

	m_skipDownloadedFiles = s->skipDownloadedFiles;
	
	if (m_downloadInProgress) {
		//If we are downloading, stop and restart the download, so it picks up the 
		//new server settings.
		stopQueue();
		m_serverManager->applySettings(s);
		startQueue();
	}
	else {
		m_serverManager->applySettings(s);
	}
}

//*****************************************************************************
//Create and terminate downloaders
//*****************************************************************************
void DownloadManager::startNewDownloader(Server *serverToUse, NntpPost *post){
	//kdDebug()<<"DownloadManager: Server to use is "<<serverToUse->hostName()<<"\n";
	serverToUse->decrementAvailableThreads();
	Downloader *dl = new Downloader(serverToUse, m_tempDir,m_postPrefix);
	m_activeDownloaders.append(dl);
	
	connect(dl,SIGNAL(message(Downloader*,int)),this,SLOT(downloaderMessage(Downloader*,int)));
	connect(dl,SIGNAL(error(Downloader*,int)),this,SLOT(downloaderError(Downloader*,int)));
	connect(dl,SIGNAL(progress(Downloader*,int,int)),this,SLOT(downloaderProgress(Downloader*,int,int)));
	
	emit threadStarted(m_activeDownloaders.count()-1, serverToUse->displayName());

	post->setState(NntpPost::DOWNLOADING);
			
	if (post->parentFile()->state()!=File::DOWNLOADING) {
		post->parentFile()->setState(File::DOWNLOADING);
		emit fileStateChanged(post->parentFile());
	}
		
	dl->downloadPost(post);
}

void DownloadManager::killDownloader(Downloader* dl) {
	emit threadFinished(m_activeDownloaders.find(dl));
	dl->server()->incrementAvailableThreads();
	m_activeDownloaders.remove(dl);
	dl->deleteLater();
}

//*****************************************************************************
//These functions are to respond to downloader messages
//*****************************************************************************
void DownloadManager::downloaderMessage(Downloader* dl, int msg) {
	switch (msg) {
		case Downloader::CONNECTED:
			emit threadConnected(m_activeDownloaders.find(dl)); 
			break;
		case Downloader::AUTHED:
			break;
		case Downloader::CHANGED_GROUP:
			emit threadChangedGroup(m_activeDownloaders.find(dl));
			break;
		case Downloader::REQUESTING:
		case Downloader::DOWNLOADING:
			emit threadDownloadStarted(m_activeDownloaders.find(dl), QString::number(dl->post()->sequenceNumber())+
				"/"+QString::number(dl->post()->parentFile()->totalSegments())+i18n(" of ")+dl->post()->
				parentFile()->filename());
			break;
		case Downloader::COMPLETED:
			dl->post()->setState(NntpPost::COMPLETE);
			downloaderCompleted(dl);
			break;
	}
}

void DownloadManager::downloaderProgress(Downloader *dl, int type, int data) {
	switch (type) {
		case Downloader::SPEED:
			emit threadSpeed(m_activeDownloaders.find(dl), data);
			break;
		case Downloader::PERCENT:
			emit threadDownloading(m_activeDownloaders.find(dl),data);
			break;
	}
}

void DownloadManager::downloaderError(Downloader* downloader,int err) {
	NntpPost *failedPost = downloader->post();
	File *failedFile = NULL;
	if (failedPost) {
		failedFile = failedPost->parentFile();
		failedPost->setState(NntpPost::FAILED);
	}

	Server *failedServer = downloader->server();
    	NntpPost *nextPost = failedServer->nextQueuedPost();
     	//If this server has any posts in its' queue, download the first.
	//Otherwise, if we're a priority 1 server - get the next queued post and start downloading
	//it.
	if (!nextPost && failedServer->priority()<2) nextPost = getNextPostInQueue();

	switch (err) {
		case Downloader::ERR_NO_POST:
		case Downloader::ERR_NO_GROUP:
			//This downloader is already connected - it just found this post/group
			//was not on the server. Use it to get the next post, and try to get
			//the missing post from the next server.
			if (nextPost)  {
				nextPost->setState(NntpPost::DOWNLOADING);
				downloader->downloadPost(nextPost);
				if (nextPost->parentFile()->state()!=File::DOWNLOADING) {
					nextPost->parentFile()->setState(File::DOWNLOADING);
					emit fileStateChanged(nextPost->parentFile());
				}
			}
			//If there are any more servers we can try for this post, try them
		    	
			if (failedPost) {
				Server *nextServer = m_serverManager->getNextServerForPost(failedPost, failedServer);
				if (nextServer) {
					//If the next server is free, download immediately
					//else put this post in its queue.
					if (nextServer->availableThreads()>0) {
						failedPost->setState(NntpPost::DOWNLOADING);
						startNewDownloader(nextServer, failedPost);
						kdDebug()<<"Downloading failed post from new server "<<nextServer->displayName()<<"\n";
					}
					else  {
						failedPost->setState(NntpPost::QUEUED);
						nextServer->queuePost(failedPost);
						kdDebug()<<"Queueing failed post to server "<<nextServer->displayName()<<"\n";
					}
				}
				else {
					kdDebug()<<"Could not find post "<<failedPost->title()<<" on any servers. Marking it as missing\n";
					failedPost->setState(NntpPost::MISSING);
				}
			}
			break;
		
		//Connection/timeout related errors
		case Downloader::ERR_DNS:
		case Downloader::ERR_CONN_REFUSED:
			//Non-transient error - don't use this server any more
			killDownloader(downloader);

			//Don't use high priority servers for a server failure, otherwise
			//we could end up eating up our transfer limits to them.
			//Only use them if a post/group isnt on the lower rated ones.
			if (err == Downloader::ERR_DNS) emit error(SERVER_ERR_DNS, failedServer->displayName());
			else if (err == Downloader::ERR_CONN_REFUSED) emit error(SERVER_ERR_CONN_REFUSED, failedServer->displayName());	
			break;
		case Downloader::ERR_TIMEOUT:
			//Keep retrying same server, same post.
			killDownloader(downloader);
			emit error(SERVER_ERR_TIMEOUT, failedServer->displayName());	
			startNewDownloader(failedServer, failedPost);
			break;

		//Bug - all error codes should be handled above.
		default:
			emit error(SERVER_ERR_UNKNOWN, failedServer->displayName());
			killDownloader(downloader);
			downloader=NULL;
	}

	if (failedFile && failedFile->readyToDecode()) {
		fileDecodeQueue.append(failedFile);
		QTimer::singleShot(0, this, SLOT(decodeFile()));
	}

	//If we have no active downloaders, the queue is complete.
	if (m_activeDownloaders.count()==0) {
		m_downloadInProgress=false;
		emit complete();
	}
}

void DownloadManager::downloaderCompleted(Downloader* completedDl) {
	File *file = completedDl->post()->parentFile();
	NntpPost *completedPost = completedDl->post();
	Server *server = completedDl->server();

	//kdDebug()<<"DownloadManager: Download completed OK for segment "<<
		//completedDl->post()->sequenceNumber()<<"of file "<< 
		//file->filename()<<"\n";

	if (completedPost->sequenceNumber()==1 && !file->existsOnDisk(m_destDir)) {
		//This is the first post of the file, and it does not already exist on disk.
		//Decode the first post to get the encoded filename, and see if that already exists.	
		Decoder *decoder = getDecoderForPost(QStringList(m_tempDir+"/"+m_postPrefix+completedPost->title()));
		if (decoder) {
			QString encName = decoder->encodedFilename();
			if (file->filename()!=encName) {
				file->setfilename(encName);
				emit fileStateChanged(file);
			
				if (file->existsOnDisk(m_destDir) && m_skipDownloadedFiles) {

					file->setState(File::EXISTS);
					file->deletePostFiles(m_tempDir, m_postPrefix);
					
					emit fileStateChanged(file);
					//Stop and restart queue, for speed.
					//NB: stopQueue deletes this downloader.
					stopQueue();
					startQueue();

					delete decoder;
					return;
				}
			}
			delete decoder;
		}
	}
	if (file->readyToDecode()) {
		fileDecodeQueue.append(file);
		QTimer::singleShot(0, this, SLOT(decodeFile()));
	}

	NntpPost *nextPost = server->nextQueuedPost();
	//If this server has any posts in it's queue, download the first.
	//Otherwise, if we're a priority 1 server, get the next and start downloading 
	//it.
	if (!nextPost && server->priority()<2) nextPost = getNextPostInQueue();
	
	//kdDebug()<<"DownloadManager: Download completed OK for segment "<<
	
	//Decoding is now done.  If this is not a priority 0/1 server, or we
	//are out of posts, kill it..
	if (!nextPost) {
//		kdDebug()<<"DownloadManager: Reaping priority "<<completedDl->server()->priority()<<" thread\n";
		killDownloader(completedDl);
	}
	else {
		//kdDebug()<<"DownloadManager: Instructing existing thread to download segment "
		//<<post->sequenceNumber()<<" of file "<<post->parentFile()->filename()<<"\n";
	
		//if this post is the first of a file, mark the file as downloading
		nextPost->setState(NntpPost::DOWNLOADING);
		if (nextPost->parentFile()->state()!=File::DOWNLOADING) {
			nextPost->parentFile()->setState(File::DOWNLOADING);
			emit fileStateChanged(nextPost->parentFile());
		}
		completedDl->downloadPost(nextPost);
	}
	
	//If we have no active downloaders, the queue is complete.
	if (m_activeDownloaders.count()==0) {
		m_downloadInProgress=false;
		emit complete();
	}
}

void DownloadManager::decodeFile() {

	if (fileDecodeQueue.isEmpty()) {
		kdDebug()<<"Asked to decode a file, but queue is empty. Aborting\n";
		return;
	}

	QValueList<File*>::Iterator flIt = fileDecodeQueue.begin();
	QGuardedPtr<File> file = *flIt;
	fileDecodeQueue.remove(flIt);
	
	//If the post we just completed is the last part we needed from this file,
	//try to decode it.
	//kdDebug()<<"DownloadManager: Completed downloading file "+file->filename()<<"\n";
	file->setState(File::DECODING);
	emit fileStateChanged(file);
	
	QStringList partList;
	QValueList<NntpPost*> posts = file->posts();
	QValueListIterator<NntpPost*> end = posts.end();	
	for (QValueListIterator<NntpPost*> p = posts.begin(); p!=end; p++) {
		partList.append(m_tempDir+"/"+m_postPrefix+(*p)->title());
	}
	
	Decoder *decoder = getDecoderForPost(partList);

	QGuardedPtr<File> fileGuarded = file;

	if (!decoder) {
		kdDebug()<<"Unrecognised post format for file\n";
		file->setState(File::DECODE_ERROR);
	}
	else {
		bool decodedOk=true;
		while (!decoder->decodingComplete()) {
			if (decoder->decodeNextPart() != Decoder::ERR_OK) {
				kdDebug()<<"Downloadmanager: Failed to decode post!\n";
				decodedOk=false;
			}
			//Process outstanding events, particularly network ones!
			kapp->processEvents();
			//If we got deleted by one of these events while decoding, return.
			if (!fileGuarded) {
				delete decoder;
				return;
			}
		}
		delete decoder;

		if (decodedOk) file->setState(File::COMPLETE);
		else file->setState(File::DECODE_ERROR);
	}
	
	file->deletePostFiles(m_tempDir, m_postPrefix);
	    
	emit fileStateChanged(file);
}


NntpPost* DownloadManager::getNextPostInQueue() {
	//Check over the current files and see if any posts are needed.
	if (m_currentFile.isNull()) {
		File *f;
		emit needFile(&f);
		m_currentFile = f;
	}

	while (!m_currentFile.isNull()) {
		if (m_currentFile->existsOnDisk(m_destDir) && m_skipDownloadedFiles) {	
			m_currentFile->setState(File::EXISTS);
			emit fileStateChanged(m_currentFile);
		}
		else {
			NntpPost *p = m_currentFile->nextNeededPost();
			if (p) {
				return p;
			}
		}
		File *f=NULL;
		emit needFile (&f);
		m_currentFile = f;
	}
	
	return NULL;
}

//*****************************************************************************
//These functions are to respond to user requests to manipulate the queue
//*****************************************************************************
void DownloadManager::startQueue() {
	QDir t(m_tempDir);
	QDir d(m_destDir);
	if (m_downloadInProgress) {
		//kdDebug()<<"Ignoring request to start queue - already downloading.\n";
	}
	else if (!t.exists()) {
		emit error(TEMP_DIR_NOT_WRITABLE,m_destDir);
	}
	else if (!d.exists()) {
		emit error(DEST_DIR_NOT_WRITABLE,m_destDir);
	} 
	else {
		bool serverAvailable=false;
		
		for (Server *server = m_serverManager->getNextServer(NULL); server; 
				server = m_serverManager->getNextServer(server)) {
			//Start off max threads for each level 0/1 priority server
			if (server->priority()<2 && server->enabled()) {
				while (server->availableThreads()>0) {
					serverAvailable=true;
					NntpPost *p = getNextPostInQueue();
					if (p) { 
						startNewDownloader(server, p);
						m_downloadInProgress=true;
					}
					else return;
				}
			}
		}	
		if (!serverAvailable) {
			emit error(SERVER_ERR_NONE_AVAILABLE,"");
		}
	}
}

void DownloadManager::stopQueue() {
	//Kill all the active downloader threads
	if (m_downloadInProgress) {
		for (Downloader *dl = m_activeDownloaders.first(); !m_activeDownloaders.isEmpty(); dl = m_activeDownloaders.first()) {
			//Requeue the posts that are currently downloading
			if (dl->post()) dl->post()->setState(NntpPost::WAITING);
			killDownloader(dl);
		}
		m_currentFile=NULL;
		m_downloadInProgress=false;
	}

	/*
	for (Server *server = m_serverManager->getNextServer(NULL); server; 
			server = m_serverManager->getNextServer(server)) {
		QValueListIterator<NntpPost*> end = server->m_postQueue.end();
		
		for (QValueListIterator<NntpPost*> it = server->m_postQueue.begin();
				it!=end; it = server->m_postQueue.begin()) {
			(*it)->setState(NntpPost::WAITING);
			server->m_postQueue.remove(it);
		}
	}
	*/
}

bool DownloadManager::downloadInProgress() {
	return m_downloadInProgress;
}
 //*****************************************************************************
 //Decoder helper
//*****************************************************************************
Decoder* DownloadManager::getDecoderForPost(QStringList posts) {
	Decoder *decoder;
	
	//Try yy
	decoder = new yyDecoder(posts,m_destDir);
	if (decoder->isDecodable()) {
		return decoder;
	}

	delete decoder;

	//Try uu
	decoder  = new uuDecoder(posts,m_destDir);
	if (decoder->isDecodable()) {
		return decoder;
	}

	delete decoder;

	//Aah, we cannot decode this post.
	return NULL;
}
	
